/*
** Copyright 2001, Michael Pfeiffer	
*/

#include "Application.h"
#include <ctype.h>
#include <stdio.h>
#include <be/interface/MenuItem.h>
#include <be/interface/Menu.h>
#include <be/interface/Alert.h>
#include <layout-all.h>
#include "detect.h"
#include "TestWindow.h"

#define APPLICATION "Hertz"
#define VERSION "1.0"

const char* configFile = "/boot/home/config/settings/kernel/drivers/gd_wrapper";
const char* fileHeader = 
"# This file is program generated, don't edit it.\n"
"# Use the preferences application \"Hertz\" instead.\n"
"# Graphics Driver Wrapper Settings File\n"
"# ~/config/settings/kernel/drivers/gd_wrapper\n"
"\n"
"# Uncomment next line to disable driver at boot time:\n";

const char* fileRefreshrate = 
"\n# Comment next line to disable changing the refreshrate for fullscreen modes:\n";

const char* fileModes  = 
"\n# Refresh rates for fullscreen applications\n"
"# Valid values are between 60 and 100\n";

static const char *configModes[] = {
	"refreshrate640x480",
	"refreshrate800x600",
	"refreshrate1024x768",
	"refreshrate1152x900",
	"refreshrate1280x1024",
	"refreshrate1600x1200",
};

static const char *modes[] = {
	"640 x 480",
	"800 x 600",
	"1024 x 768",
	"1152 x 900",
	"1280 x 1024",
	"1600 x 1200",
	NULL
};

static uint32 space[] = {
	B_8_BIT_640x480,
	B_8_BIT_800x600,
	B_8_BIT_1024x768,
	B_8_BIT_1152x900,
	B_8_BIT_1280x1024,
	B_8_BIT_1600x1200
};

/*
static uint32 space[] = {
	B_16_BIT_640x480,
	B_16_BIT_800x600,
	B_16_BIT_1024x768,
	B_16_BIT_1152x900,
	B_16_BIT_1280x1024,
	B_16_BIT_1600x1200
};
*/

/*
static uint32 space[] = {
	B_32_BIT_640x480,
	B_32_BIT_800x600,
	B_32_BIT_1024x768,
	B_32_BIT_1152x900,
	B_32_BIT_1280x1024,
	B_32_BIT_1600x1200
};
*/

// Note: GDW allows only refresh rates between 60 and 100 Hz
static const char *rrNames[] = {
	"60 Hz",
	"72 Hz",
	"75 Hz",
	"80 Hz",
	"85 Hz",
	"90 Hz",
	"95 Hz",
	"100 Hz",
//	"110 Hz", 
//	"120 Hz",
	NULL
};

static const int refreshRate[] = {
	60, 72, 75, 80, 85, 90, 95, 100, /* 110, 120, */ -1
};

static minimax _min(0, 0, 10, 30, 1);

void AppWindow::FillMenu(BMenu* menu, const char* labels[], uint32 what) {
	for (int i = 0; labels[i]; i ++) {
		BMessage* msg = new BMessage(what);
		msg->AddInt8("index", i);
		menu->AddItem(new BMenuItem((const char*)labels[i], msg, 0));
	}
}

bool find_string(char* b, int &i, char **s){
	while(b[i]) {
		if (b[i] == '#' || !isalnum(b[i])) { // skip to end of line
			while (b[i] && b[i] != '\n') i++;
			if (b[i] == '\n') i++;
		} else {
			*s = &b[i]; i++;
			// skip to whitespace
			while (b[i] && isalnum(b[i])) i++;
			if (b[i]) {
				b[i] = 0; // terminate string s
				i++;
			}
			// skip whitespaces
			while (b[i] && !isalnum(b[i])) i++;
			return true;
		}
	}
	return false;
}

void AppWindow::LoadSettings() {
	BFile file(configFile, B_READ_ONLY);
	// set default values
	for (int i = 0; i < _no_of_modes; i ++) {
		_initial.refreshrate[i] = 60;
	}

	_initial.driver_enabled = true;
	_initial.change_refreshrate = false;
	
	off_t size;
	// parse file
	if (file.InitCheck() == B_OK && file.GetSize(&size) == B_OK && size > 0) {
		char* buffer = new char[size+1];
		if (file.Read(buffer, size) == size) {
			int index = 0;
			char* s;
			buffer[size] = 0;
			while (find_string(buffer, index, &s)) {
				if (strcmp(s, "disabledriver") == 0) {
					_initial.driver_enabled = false;
				} else if (strcmp(s, "changerefreshrate") == 0) {
					_initial.change_refreshrate = true;
				} else {
					for (int i = 0; i < _no_of_modes; i++) {
						if (strcmp(s, configModes[i]) == 0) {
							if (find_string(buffer, index, &s)) {
								int rf = atoi(s);
								if (rf < 60) rf = 60;
								else if (rf > 100) rf = 100;
								_initial.refreshrate[i] = rf;
							}
							break;
						}
					}
				}
			}
		}
		delete[] buffer;
	}

	_settings = _initial;
}

void AppWindow::SaveSettings(struct settings* s) {
	BFile file(configFile, B_CREATE_FILE | B_ERASE_FILE | B_WRITE_ONLY);
	if (file.InitCheck() == B_OK) {
		file.Write(fileHeader, strlen(fileHeader));
		if (s->driver_enabled) file.Write("#", 1);
		file.Write("disabledriver\n", 14);
		
		file.Write(fileRefreshrate, strlen(fileRefreshrate));
		if (!s->change_refreshrate) file.Write("#", 1);
		file.Write("changerefreshrate\n", 18);
		
		file.Write(fileModes, strlen(fileModes));
		for (int i = 0; i < _no_of_modes; i++) {
			file.Write(configModes[i], strlen(configModes[i]));
			file.Write(" ", 1);
			char buffer[10];
			sprintf(buffer, "%d", s->refreshrate[i]);
			file.Write(buffer, strlen(buffer));
			file.Write("\n", 1);
		}
	}
}

AppWindow::AppWindow(BRect aRect) 
	: MWindow(aRect, APPLICATION, 
		B_TITLED_WINDOW_LOOK, 
		B_NORMAL_WINDOW_FEEL, 
		B_NOT_ZOOMABLE | B_NOT_RESIZABLE) {

	LoadSettings();

	bool gdw_found = gdw_detect(NULL) == B_OK;

	MPopup *res, *ref;
	MView *mView = new VGroup(
		new Space(minimax(3, 3, 3, 3)),
		new MStringView("USE AT YOUR OWN RISK!", B_ALIGN_CENTER),
		new MStringView("TOO HIGH REFRESH RATES MIGHT", B_ALIGN_CENTER),
		new MStringView("DESTROY YOUR MONITOR!", B_ALIGN_CENTER),
		new Space(minimax(8, 8, 8, 8)),
		new MBorder(M_LABELED_BORDER, 5, "Full Screen Display Mode",
			new HGroup(
				new VGroup(
					res = new MPopup("Resolution:", new BMessage(RESOLUTION_CHANGED), NULL, 0),
					ref = new MPopup("Refresh Rate:", new BMessage(REFRESH_RATE_CHANGED), NULL, 0),
				0),
				new Space(/*minimax(5, 5, 5, 5)*/),
				new VGroup(
					new MButton("Test", new BMessage(TEST_CLICKED), NULL, minimax(-1, -1, -1, -1)),
					new MButton("About", new BMessage(B_ABOUT_REQUESTED), NULL, minimax(-1, -1, -1, -1)),
				0),
				new Space(minimax(5, 5, 5, 5)),
			0)
		),
		new Space(minimax(8, 8, 8, 8)),
		_driverCB = new MCheckBox("Enable Driver (reboot required)", new BMessage(DRIVER_ENABLED_CHANGED), NULL, _settings.driver_enabled),	
		_refCB    = new MCheckBox("Change Refresh Rate", new BMessage(CHANGE_REF_CHANGED), NULL, _settings.change_refreshrate),	
/*
		new HGroup(
			new MButton("Default", new BMessage(DEFAULT_CLICKED), NULL, _min), 
			new MButton("Undo", new BMessage(UNDO_CLICKED), NULL, _min), 
			new MButton("Apply", new BMessage(APPLY_CLICKED), NULL, _min),
		0),
*/
		new Space(minimax(5, 5, 5, 5)),
		new MStringView(gdw_found ? "Status: Graphics Driver Wrapper not installed or inactive!" : "Status: Graphics Driver Wrapper active."),
		new Space(minimax(3, 3, 3, 3)),
	0);
	FillMenu(_resMenu = res->Menu(), modes, RESOLUTION_CHANGED);
	_resMenu->SetRadioMode(true);
	FillMenu(_refMenu = ref->Menu(), rrNames, REFRESH_RATE_CHANGED);
	_refMenu->SetRadioMode(true);
	mView = new HGroup(
		new Space(minimax(3, 3, 3, 3)),
		mView,
		new Space(minimax(3, 3, 3, 3)),
		0);
	AddChild(dynamic_cast<BView*>(mView));
	SelectResMenu(_resIndex = 0);
	Show();
}

void AppWindow::SelectMenu(BMenu* menu, int index) {
	BMenuItem* item = menu->ItemAt(index);
	if (item) {
		item->SetMarked(true);
	}
}

void AppWindow::SelectResMenu(int i) {
	SelectMenu(_resMenu, i);
	int rf = _settings.refreshrate[i];
	int min = 200;
	for (i = 0; refreshRate[i] != -1; i++) {
		int m = abs(refreshRate[i] - rf);
		if (m < min) {
			min = m;
		} else {
			break;
		}
	}
	SelectMenu(_refMenu, i-1);
}

void AppWindow::OnResolutionChanged(BMessage* msg) {
	int8 i;
	if (B_OK == msg->FindInt8("index", &i)) {
		_resIndex = i;
		SelectResMenu(i);
	}
}

void AppWindow::OnRefreshRateChanged(BMessage* msg) {
	int8 i;
	if (B_OK == msg->FindInt8("index", &i)) {
		_settings.refreshrate[_resIndex] = refreshRate[i];
		SaveSettings(&_settings);
	}
}

static bool IsOn(BMessage *msg) {
	void *ptr;
	if (B_OK == msg->FindPointer("source", &ptr)) {
		return ((BControl*)ptr)->Value() == B_CONTROL_ON;
	}
	return false;
}

void AppWindow::OnDriverEnabledChanged(BMessage* msg) {
	_settings.driver_enabled = IsOn(msg);
		SaveSettings(&_settings);
}

void AppWindow::OnChangeRefChanged(BMessage* msg) {
	_settings.change_refreshrate = IsOn(msg);
		SaveSettings(&_settings);
}

void AppWindow::OnTestClicked(BMessage* msg) {
	static bool alertShown = false;

	if (!alertShown) {
		alertShown = true;
		BAlert *alert = new BAlert(
			"Reminder", 
			"A white rectangle (marking the screen borders) is shown in a full screen window"
			" after clicking the OK button.\n"
			"Hit Escape key to return from fullscreen mode!",
			"Ok", NULL, NULL,
			B_WIDTH_AS_USUAL,
			B_WARNING_ALERT);
		alert->SetShortcut(0, B_ESCAPE);
		alert->Go();
	}


	status_t error;
	// save current settings
	TestWindow* win = new TestWindow(space[_resIndex], &error);
	if (error != B_OK) {
		win->Quit(); 
		BAlert *alert = new BAlert(
			"Error", 
			"Could not switch into full screen mode!",
			"Ok", NULL, NULL,
			B_WIDTH_AS_USUAL,
			B_WARNING_ALERT);
		alert->SetShortcut(0, B_ESCAPE);
		alert->Go();
	}
	// save initial settings
}


void AppWindow::OnDefaultClicked(BMessage* msg) {
	for (int i = 0; i < _no_of_modes; i++) _settings.refreshrate[i] = 60;
	SelectResMenu(_resIndex);
}


void AppWindow::OnUndoClicked(BMessage* msg) {
	_settings = _initial;
	SelectResMenu(_resIndex);
	_driverCB->SetValue(_settings.driver_enabled);
	_refCB->SetValue(_settings.change_refreshrate);
}


void AppWindow::OnApplyClicked(BMessage* msg) {
	SaveSettings(&_settings);
}


void AppWindow::MessageReceived(BMessage *message) {
	message->PrintToStream();
	
	switch(message->what) {
	case RESOLUTION_CHANGED:     OnResolutionChanged(message); break;
	case REFRESH_RATE_CHANGED:   OnRefreshRateChanged(message); break;
	case DRIVER_ENABLED_CHANGED: OnDriverEnabledChanged(message); break;
	case CHANGE_REF_CHANGED:     OnChangeRefChanged(message); break;
	case TEST_CLICKED:           OnTestClicked(message); break;
	case DEFAULT_CLICKED:        OnDefaultClicked(message); break;
	case UNDO_CLICKED:           OnUndoClicked(message); break;
	case APPLY_CLICKED:          OnApplyClicked(message); break;
	case B_ABOUT_REQUESTED:
		AboutRequested();
		break;
	default:
		BWindow::MessageReceived(message);
	}
}


bool AppWindow::QuitRequested() {
	be_app->PostMessage(B_QUIT_REQUESTED);
	return(true);
}

void AppWindow::AboutRequested() {
	BAlert *about = new BAlert(APPLICATION, 
		APPLICATION " " VERSION "\nThis program is freeware under GPL.\n\n"
		"Written 2001.\n\n"
		"By Michael Pfeiffer.\n\n"
		"EMail: michael.pfeiffer@utanet.at.","Close");
		about->Go();
}

App::App() : BApplication("application/x-vnd.mp-hertz") {
	BRect aRect;
	// set up a rectangle and instantiate a new window
	aRect.Set(100, 80, 210, 180);
	window = NULL;
	window = new AppWindow(aRect);		
}

int main(int argc, char *argv[]) { 
	be_app = NULL;
	App app;
	be_app->Run();
	return 0;
}